import logging
from flask import request
from flask_jwt_extended import jwt_required
from flask_restplus import Namespace, Resource

from portal.rest import rest_helpers
from portal.rest.schemas import group_import_schema, group_schema, groups_schema, projects_schema, \
    user_list_update_schema, users_schema
from portal.service import auth, general
from portal.service.errors import ForbiddenError
from portal.service.groups import add_project_to_group, add_single_user_to_group, create_group, \
    delete_group, find_projects_in_group, find_users_in_group_by_role, import_group, list_groups, \
    remove_project_from_group, remove_single_user_from_group, update_group, \
    update_user_group_membership
from portal.service.permissions import check_client, require_client

groups_namespace = Namespace('')

log = logging.getLogger(__name__)


@groups_namespace.route('/courses/<string:cid>/groups')
@groups_namespace.param('cid', 'Course id')
@groups_namespace.response(404, 'Course not found')
class GroupsList(Resource):
    @jwt_required
    # @groups_namespace.response(200, 'List of all groups', model=groups_schema)
    def get(self, cid: str):
        client = auth.find_client()
        course = general.find_course(cid)

        return groups_schema.dump(list_groups(course, client=client))

    @jwt_required
    # @groups_namespace.response(201, 'Group created', model=group_schema)
    @groups_namespace.response(403, 'Not allowed to create group')
    def post(self, cid: str):
        client = auth.find_client()
        course = general.find_course(cid)
        # authorization
        require_client(client=client, course=course,
                       permissions=['update_course'])

        data = rest_helpers.parse_request_data(
            group_schema, action='create', resource='group'
        )

        new_group = create_group(course, **data)
        return group_schema.dump(new_group)[0], 201


@groups_namespace.route('/courses/<string:cid>/groups/<string:gid>')
@groups_namespace.param('cid', 'Course id')
@groups_namespace.param('gid', 'Group id')
@groups_namespace.response(404, 'Course not found')
@groups_namespace.response(404, 'Group not found')
class GroupResource(Resource):
    @jwt_required
    # @groups_namespace.response(201, 'Group', model=group_schema)
    @groups_namespace.response(403, 'Not allowed to get group')
    def get(self, cid: str, gid: str):
        client = auth.find_client()
        course = general.find_course(cid)
        # authorization
        permissions = ['view_course_full', 'view_course_limited']
        require_client(client=client, course=course, permissions=permissions)

        group = general.find_group(course, gid)
        return group_schema.dump(group)

    @jwt_required
    @groups_namespace.response(204, 'Group deleted')
    @groups_namespace.response(403, 'Not allowed to delete group')
    def delete(self, cid: str, gid: str):
        client = auth.find_client()
        course = general.find_course(cid)
        # authorization
        require_client(client=client, course=course,
                       permissions=['update_course'])

        group = general.find_group(course, gid)
        delete_group(group)
        return '', 204

    @jwt_required
    @groups_namespace.response(204, 'Group updated')
    @groups_namespace.response(403, 'Not allowed to update group')
    def put(self, cid: str, gid: str):
        client = auth.find_client()
        course = general.find_course(cid)
        # authorization
        permissions = ['write_groups', 'update_course']
        require_client(client=client, course=course, permissions=permissions)

        data = rest_helpers.parse_request_data(
            group_schema, action='update', resource='group'
        )

        group = general.find_group(course, gid)
        update_group(group, data)
        return '', 204


@groups_namespace.route('/courses/<string:cid>/groups/<string:gid>/users')
@groups_namespace.param('cid', 'Course id')
@groups_namespace.param('gid', 'Group id')
class GroupUsersList(Resource):
    @jwt_required
    # @groups_namespace.response(200, 'List of users in the group', model=users_schema)
    @groups_namespace.response(403, 'Not allowed to get users in the group')
    def get(self, cid: str, gid: str):
        client = auth.find_client()
        course = general.find_course(cid)
        role_id = request.args.get('role')

        group = general.find_group(course, gid)
        # authorization
        if not check_client(client=client, course=course, permissions='view_course_full') or \
                (check_client(client=client, course=course, permissions=['view_course_limited'])
                 and client in group.users):
            raise ForbiddenError(uid=client.id)

        users = find_users_in_group_by_role(course, group, role_id)
        return users_schema.dump(users)

    @jwt_required
    @groups_namespace.response(204, 'User\'s list updated')
    @groups_namespace.response(403, 'Not allowed to add users to the group')
    def put(self, cid: str, gid: str):
        client = auth.find_client()
        course = general.find_course(cid)
        # authorization
        permissions = ['write_groups', 'update_course']
        require_client(client=client, course=course, permissions=permissions)

        data = rest_helpers.parse_request_data(
            user_list_update_schema, action='update', resource='group_membership'
        )

        group = general.find_group(course, gid)
        # everything from users_add is added, THEN everything from users_remove
        # is removed
        update_user_group_membership(group, data)
        return '', 204


@groups_namespace.route(
    '/courses/<string:cid>/groups/<string:gid>/users/<string:uid>')
@groups_namespace.param('cid', 'Course id')
@groups_namespace.param('gid', 'Group id')
@groups_namespace.param('uid', 'User id')
@groups_namespace.response(404, 'Course not found')
@groups_namespace.response(404, 'Group not found')
@groups_namespace.response(404, 'User not found')
class GroupAddOrDeleteSingleUser(Resource):
    @jwt_required
    @groups_namespace.response(204, 'User\'s list updated')
    @groups_namespace.response(403, 'Not allowed to add user to the group')
    def put(self, cid: str, gid: str, uid: str):
        client = auth.find_client()
        course = general.find_course(cid)

        # authorization
        permissions = ['write_groups', 'update_course']
        require_client(client=client, course=course, permissions=permissions)

        group = general.find_group(course, gid)
        user = general.find_user(uid)
        add_single_user_to_group(group, user)
        return '', 204

    @jwt_required
    @groups_namespace.response(204, 'User\'s list updated')
    @groups_namespace.response(403, 'Not allowed to delete users to the group')
    def delete(self, cid: str, gid: str, uid: str):
        client = auth.find_client()
        course = general.find_course(cid)

        # authorization
        permissions = ['write_groups', 'update_course']
        require_client(client=client, course=course, permissions=permissions)

        group = general.find_group(course, gid)
        user = general.find_user(uid)
        remove_single_user_from_group(group, user)
        return '', 204


@groups_namespace.route('/courses/<string:cid>/groups/<string:gid>/projects')
@groups_namespace.param('cid', 'Course id')
@groups_namespace.param('gid', 'Group id')
@groups_namespace.response(404, 'Course not found')
@groups_namespace.response(404, 'Group not found')
class GroupProjectsList(Resource):
    @jwt_required
    def get(self, cid: str, gid: str):
        client = auth.find_client()
        course = general.find_course(cid)
        group = general.find_group(course, gid)
        # authorization
        if not check_client(client=client, course=course, permissions=['view_course_full']) or \
                (check_client(client=client, course=course, permissions=['view_course_limited'])
                 and client in group.users):
            raise ForbiddenError(uid=client.id)
        projects = find_projects_in_group(group)
        return projects_schema.dump(projects)


@groups_namespace.route(
    '/courses/<string:cid>/groups/<string:gid>/projects/<string:pid>')
@groups_namespace.param('cid', 'Course id')
@groups_namespace.param('gid', 'Group id')
@groups_namespace.param('pid', 'Project id')
@groups_namespace.response(404, 'Course not found')
@groups_namespace.response(404, 'Group not found')
@groups_namespace.response(404, 'Project not found')
class GroupAddOrDeleteProject(Resource):
    @jwt_required
    @groups_namespace.response(204, 'Projects\'s list updated')
    @groups_namespace.response(403, 'Not allowed to add project to the group')
    def put(self, cid: str, gid: str, pid: str):
        client = auth.find_client()
        course = general.find_course(cid)

        # authorization
        permissions = ['write_groups', 'update_course']
        require_client(client=client, course=course, permissions=permissions)

        group = general.find_group(course, gid)
        project = general.find_project(course, pid)
        add_project_to_group(group, project)
        return '', 204

    @jwt_required
    @groups_namespace.response(204, 'Projects\'s list updated')
    @groups_namespace.response(
        403, 'Not allowed to delete project from the group')
    def delete(self, cid: str, gid: str, pid: str):
        client = auth.find_client()
        course = general.find_course(cid)

        # authorization
        permissions = ['write_groups', 'update_course']
        require_client(client=client, course=course, permissions=permissions)

        group = general.find_group(course, gid)
        project = general.find_project(course, pid)
        remove_project_from_group(group, project)
        return '', 204


@groups_namespace.route('/courses/<string:cid>/groups/import')
@groups_namespace.param('cid', 'Course id')
@groups_namespace.response(404, 'Course not found')
class GroupImport(Resource):
    @jwt_required
    @groups_namespace.response(201, 'Group import')
    @groups_namespace.response(404, 'Group not found')
    @groups_namespace.response(403, 'Not allowed to import to the group')
    def put(self, cid: str):
        client = auth.find_client()
        target_course = general.find_course(cid)
        # authorization
        permissions = ['write_groups', 'update_course']
        require_client(client=client, course=target_course,
                       permissions=permissions)
        data = rest_helpers.parse_request_data(
            group_import_schema, action='import', resource='group'
        )
        new_group = import_group(data, target_course)
        return group_schema.dump(new_group), 201
