"""
Submissions service
"""

import logging
from typing import Union
from celery.result import AsyncResult

from portal.async_celery import tasks
from portal.async_celery.enqueue import revoke_enqueue_task
from portal.async_celery.storage import revoke_storage_task
from portal.async_celery.tasks import clone_submission_files, upload_submission_to_storage
from portal.database.models import Component, Project, Submission, SubmissionState, User
from portal.service.errors import ForbiddenError
from portal.service.general import delete_entity, write_entity

log = logging.getLogger(__name__)


def create_submission(user: User, project: Project, file_params: dict,
                      project_params_string: str) -> Submission:
    """Creates a new submission in the database and downloads its files.

    Zip and git sources are processed differently.

    Args:
        user(User): the submission's author
        project(Project): the project the submission falls under
        file_params(dict): parameters for Storage - how and where to download files
        project_params_string(str): custom parameters specified by
            project with their values as a JSON-encoded string

    Returns(Submission): the created submission entity

    """
    # adding review at submission create is not supported
    new_submission = Submission(
        user=user, project=project, parameters=project_params_string)
    write_entity(new_submission)

    # download relevant files into Storage
    save_submission_to_storage(new_submission, file_params)
    enqueue_submission(new_submission, submission_params=file_params)
    return new_submission


def save_submission_to_storage(new_submission: Submission, file_params: dict) -> AsyncResult:
    project = new_submission.project
    file_params['whitelist'] = project.config.file_whitelist
    if file_params['source']['type'] == 'zip':
        file_params['source']['path'] = "todo"  # TODO: download zip

    task = create_save_task(file_params, new_submission)
    return task


def enqueue_submission(new_submission: Submission, submission_params: dict):
    project = new_submission.project
    time_wait = project.config.submissions_cancellation_period
    result: AsyncResult = tasks.start_processing_submission \
        .delay(new_submission.id, submission_params)


def create_save_task(file_params, new_submission) -> AsyncResult:
    result: AsyncResult = upload_submission_to_storage.delay(
        new_submission.id, file_params)
    new_submission.storage_task_id = result.task_id
    return result


def copy_submission(source: Submission, note: str) -> Submission:
    """Copies a submission. Used at resubmitting

    Args:
        source(Submission):
        note(str): Note for the submission. Typically contains a reason for the resubmit.

    Returns(Submission): Copied submission
    """
    log.info(f"[COPY] Submission {source.id} to "
             f"same project {source.project.id}")
    new_submission = Submission(user=source.user, project=source.project,
                                parameters=source.parameters)
    new_submission.note = note
    write_entity(new_submission)
    clone_submission_files.delay(
        source_id=source.id, target_id=new_submission.id)
    return new_submission


def delete_submission(submission: Submission):
    """Deletes a submission

    Args:
        submission(Submission): Submission instance
    """
    log.info(f"[DELETE] Submission {submission.id}")
    # TODO: dispatch action to clean the storage
    delete_entity(submission)


def update_submission_state(client: Union[User, Component], submission: Submission,
                            data: dict) -> Submission:
    """Updates submission's state

    Args:
        client(User): Client attempting the update
        submission(Submission): Submission instance to update
        data(dict): should contain one key, 'state' with the new value

    Returns(Submission): Updated submission
    """
    new_state = data['state']
    if isinstance(client, User) and new_state != SubmissionState.CANCELLED:
        raise ForbiddenError(uid=client.id,
                             note=f"User {client.id} cannot update "
                                  f"state to other than CANCELLED.")
    submission.change_state(new_state)
    write_entity(submission)
    log.info(f"[UPDATE] Submission state {submission.id} "
             f"by {client.id} to {submission.state}")
    return submission


def cancel_submission(submission: Submission):
    """Cancels the submission
    Args:
        submission(Submission): Submission instance

    """
    submission.state = SubmissionState.CANCELLED
    revoke_enqueue_task(submission=submission)
    revoke_storage_task(submission=submission)
    write_entity(submission)
